# (c) Copyright 2019 Trent Hauck
# All Rights Reserved
"""The base tokenizer from which other tokenizers are expected to inhereit.

The `SequenceTokenizerSettings` in particular holds common tokens and their associated ids for
common special tokens. For example `unk_token` for unknown tokens.
"""

from typing import List
from typing import Optional

from pydantic import BaseSettings
from pydantic import Field
from pydantic import root_validator


# pylint: disable=too-few-public-methods
class SequenceTokenizerSettings(BaseSettings):
    """The base tokenizer settings which inherits from pydantic.BaseSettings.

    To see the available fields, use pydantic's `schema_json`.

    ```python
    >>> print(SequenceTokenizerSettings.schema_json(indent=2))
    {
      "title": "SequenceTokenizerSettings"
      ...
    }
    """

    unk_token: Optional[str] = Field(None, env="GCGC_UNK_TOKEN")
    unk_token_id: Optional[int] = Field(None, env="GCGC_UNK_TOKEN_ID")

    bos_token: Optional[str] = Field(None, env="GCGC_BOS_TOKEN")
    bos_token_id: Optional[int] = Field(None, env="GCGC_BOS_TOKEN_ID")

    mask_token: Optional[str] = Field(None, env="GCGC_MASK_TOKEN")
    mask_token_id: Optional[int] = Field(None, env="GCGC_MASK_TOKEN_ID")

    eos_token: Optional[str] = Field(None, env="GCGC_EOS_TOKEN")
    eos_token_id: Optional[int] = Field(None, env="GCGC_EOS_TOKEN_ID")

    pad_token: Optional[str] = Field(None, env="GCGC_PAD_TOKEN")
    pad_token_id: Optional[int] = Field(None, env="GCGC_PAD_TOKEN_ID")

    max_length: Optional[int] = Field(None, env="GCGC_MAX_LENGTH")
    min_length: Optional[int] = Field(None, env="GCGC_MIN_LENGTH")

    conform_length: Optional[int] = Field(None, env="GCGC_CONFORM_LENGTH")

    # pylint: disable=no-self-argument, no-self-use
    @root_validator
    def validate_and_set_special_tokens(cls, values):
        """Check that the tokens and token id pairs have corresponding values.

        If they do not, this will set the default value.

        """
        special_token_defaults = [
            ("pad_token", "|", 0),
            ("bos_token", ">", 1),
            ("eos_token", "<", 2),
            ("unk_token", "?", 3),
        ]
        for token_name, token_char, token_id in special_token_defaults:
            passed_token = values.get(token_name)
            passed_token_id = values.get(f"{token_name}_id")

            if (passed_token is None and passed_token_id is None) or (
                passed_token is not None and passed_token_id is not None
            ):
                continue

            if passed_token is None and passed_token_id is not None:
                values[token_name] = token_char

            if passed_token is not None and passed_token_id is None:
                values[f"{token_name}_id"] = token_id

        return values

    @root_validator
    def validate_lengths(cls, values):  # pylint: disable=no-self-argument, no-self-use
        """Check the length arguments are valid."""
        max_length = values.get("max_length")
        min_length = values.get("min_length")
        conform_length = values.get("conform_length")
        pad_token = values.get("pad_token")

        if conform_length is not None and (max_length is not None or min_length is not None):
            raise ValueError(f"If conform length is not None, max and min length can't be set.")

        if max_length is not None and min_length is not None and min_length > max_length:
            raise ValueError(
                f"Min length {min_length} cannot be more than max length {max_length}."
            )

        if (conform_length is not None or min_length is not None) and pad_token is None:
            raise ValueError("Cannot set conform_length or min_length without a pad_token.")

        return values

    @property
    def special_token_ids(self) -> List[int]:
        """Return the list of token ids corresponding to special tokens."""
        return [
            x
            for x in [
                self.bos_token_id,
                self.eos_token_id,
                self.unk_token_id,
                self.pad_token_id,
                self.mask_token_id,
            ]
            if x is not None
        ]

    @property
    def special_tokens(self) -> List[str]:
        """Return the set of special tokens that are not None."""
        return [
            x
            for x in [
                self.bos_token,
                self.eos_token,
                self.unk_token,
                self.pad_token,
                self.mask_token,
            ]
            if x is not None
        ]


def _pad_token_list(tokens: List[str], pad_to: int, pad_char: str):
    """Pad the input tokens list to tokens."""
    token_len = len(tokens)

    if token_len > pad_to:
        return tokens

    new_tokens = pad_to - token_len
    return tokens + [pad_char] * new_tokens


class SequenceTokenizer:
    """The sequence tokenizer object."""

    def __init__(self, settings: Optional[SequenceTokenizerSettings] = None):
        """Init the sequence tokenizer with a set of settings."""
        self.settings = settings or SequenceTokenizerSettings()

    def get_special_tokens_mask(self, token_ids: List[int]) -> List[int]:
        """Given the input set of tokens, return a list that demarcates special character.

        ```python
        # Assuming 1 is bos_token, 2 is eos_token, and 0 is pad_token.
        >>> tokenizer.get_special_tokens_mask([1, 34, 21, 0, 0, 0, 2])
        [1, 0, 0, 1, 1, 1, 1]
        ```

        Args:
            token_ids: The list of integer tokens that may or may not be special tokens according
                to the toeknizer's settings.

        Returns:
            A list of 0s and 1s, where the value is one if the token is a special character.

        """
        return [int(token_id in self.settings.special_token_ids) for token_id in token_ids]

    def apply_length_constraints(self, tokens: List[str]) -> List[str]:
        """Apply the constraints from the settings to the passed tokens list."""
        if self.settings.conform_length:
            tokens = _pad_token_list(tokens, self.settings.conform_length, self.settings.pad_token)
            return tokens[: self.settings.conform_length]

        if self.settings.min_length:
            tokens = _pad_token_list(tokens, self.settings.min_length, self.settings.pad_token)

        if self.settings.max_length:
            tokens = tokens[: self.settings.max_length]

        return tokens



# (c) Copyright 2019 Trent Hauck
# All Rights Reserved
"""This module holds the kmer tokenizer and its settings.

The `KmerTokenizer` tokenizers incoming sequences into kmers of configurable size against a
configurable alphabet.

For example, given incoming nucleotide sequences, using settings of kmer length 3 and stride 3 they
encoded sequences will be codons (in the loose sense) with a vocabulary of size 64.

```python
>>> from gcgc.tokenizer import KmerTokenizer, KmerTokenizerSettings

>>> settings = KmerTokenizerSettings(alphabet="ATCG", kmer_length=3, kmer_stride=3)
>>> tokenizer = KmerTokenizer(settings=settings)

>>> tokenizer.encode('AAATTTCCC')
[0, 42, 63]

>>> len(tokenizer.vocab)
64
```
"""

import itertools as it
from typing import Dict
from typing import List
from typing import Optional

from pydantic import Field
from pydantic import validator

from gcgc import alphabets
from gcgc.tokenizer.base import SequenceTokenizer
from gcgc.tokenizer.base import SequenceTokenizerSettings


class KmerTokenizerSettings(SequenceTokenizerSettings):
    """The specification for the tokenizer.

    Like the baseclass, `SequenceTokenizerSettings`, the schema (and thus available fields), can be
    seen by using the `print_schema` classmethod.

    ```python
    >>> print(KmerTokenizerSettings.schema_json(indent=2))
    {
      "title": "SequenceTokenizerSettings"
      ...
    }

    """

    alphabet: str = Field("ATCG", env="GCGC_ALPHABET")
    kmer_length: int = Field(1, env="GCGC_KMER_LENGTH")
    kmer_stride: int = Field(1, env="GCGC_KMER_STRIDE")

    # pylint: disable=no-self-use, no-self-argument
    @validator("alphabet")
    def resolve_alphabet(cls, alphabet: str) -> str:
        """Resolve the alphabet if it's a named alphabet.

        Args:
            alphabet: The raw alphabet, either the sequence literal or a name of a preset alphabet.

        Returns:
            The new alphabet.

        """
        return alphabets.resolve_alphabet(alphabet)


def _create_kmer_vocab_from_token(settings: KmerTokenizerSettings) -> Dict[str, int]:
    """Create the vocab object from a list of tokens."""
    token_to_int = {}

    for token_id, token in zip(settings.special_token_ids, settings.special_tokens):
        token_to_int[token] = token_id

    token_set = [
        "".join(kmer) for kmer in it.product(list(settings.alphabet), repeat=settings.kmer_length)
    ]

    starting_value = 0
    for token in token_set:
        while starting_value in settings.special_token_ids:
            starting_value += 1

        token_to_int[token] = starting_value
        starting_value += 1

    return token_to_int


class KmerTokenizer(SequenceTokenizer):
    """The Kmer Tokenizer that encodes sequences into chunked kmers."""

    def __init__(self, settings: Optional[KmerTokenizerSettings] = None):
        """Init the SequenceTokenizer class.

        Args:
            settings: The settings for the tokenizer.

        """
        self.settings = settings or KmerTokenizerSettings()
        super().__init__(settings)

        self.vocab = _create_kmer_vocab_from_token(self.settings)

    def _kmer_n(self, seq: str) -> List[str]:
        seq_len = len(seq)
        iterations = seq_len - self.settings.kmer_length + 1
        kmers = []

        for i in range(0, iterations, self.settings.kmer_stride):
            kmer = seq[i : i + self.settings.kmer_length]
            kmers.append(kmer)

        return kmers

    def encode(self, seq: str, add_unknown: bool = False) -> List[int]:
        """Encode the underlying sequence into a list of tokens ids.

        Args:
            seq: The incoming sequence to encode.
            add_unknown: If True, add the unknown token rather than throwing an out of vocabulary
                error.

        Returns:
            A list of the encoded tokens.

        """
        encoded = []

        for letter in self.encode_as_tokens(seq):
            try:
                encoded.append(self.vocab[letter])
            except KeyError:
                if add_unknown:
                    encoded.append(self.settings.unk_token_id)
                else:
                    raise

        return encoded

    def encode_as_tokens(self, seq: str) -> List[str]:
        """Tokenize the sequence into a list of tokens.

        Args:
            seq: The sequence to encode.

        Returns:
            The list of strs that are the tokens.

        """
        seq_len = len(seq)

        if seq_len < self.settings.kmer_length:
            raise ValueError(
                f"seq length {seq_len} cannot be less than the kmer "
                "size {self.settings.kmer_length}"
            )

        if self.settings.kmer_length == 1:
            kmer_list = list(seq)
        else:
            kmer_list = self._kmer_n(seq)

        if self.settings.bos_token:
            kmer_list = [self.settings.bos_token] + kmer_list

        if self.settings.eos_token:
            kmer_list = kmer_list + [self.settings.eos_token]

        return super().apply_length_constraints(kmer_list)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Top-level GCGC module."""

import warnings as _warnings

from gcgc import tokenizer
from gcgc.cli import cli
from gcgc.tokenizer import KmerTokenizer
from gcgc.tokenizer import KmerTokenizerSettings
from gcgc.tokenizer import SequenceTokenizer
from gcgc.tokenizer import SequenceTokenizerSettings

__version__ = "0.12.1-dev.5"

_warnings.simplefilter(action="ignore", category=PendingDeprecationWarning)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Test the sentence models."""

import json
import random
from pathlib import Path

import pytest

from gcgc import cli_utils
from gcgc.alphabets import resolve_alphabet
from gcgc.tests import fixtures
from gcgc.tokenizer import sentence_piece_tokenizer


@pytest.mark.parametrize("alphabet", ["protein", "unambiguous_dna"])
def test_train_sentence_piece(tmp_path, alphabet):
    """Test fitting the sentence piece model."""
    alphabet_list = resolve_alphabet(alphabet, require=True)

    sequences = []
    for _ in range(500):
        new_seq = []
        for _ in range(500):
            new_seq.append(random.choice(list(alphabet_list)))

        sequences.append("".join(new_seq))

    model_prefix = tmp_path / "model"

    settings = sentence_piece_tokenizer.BioSequencePieceSettings(
        model_prefix=model_prefix, vocab_size=50, bos_token=">", eos_token="<", pad_token="|"
    )
    sp_tokenizer = sentence_piece_tokenizer.BioSequencePiece(settings)
    sp_tokenizer.fit_on_list(sequences)

    test_case = "".join(random.choice(alphabet_list) for _ in range(20))

    tokenized = sp_tokenizer.encode_as_tokens(test_case)
    assert all([isinstance(x, str) for x in tokenized])

    ids = sp_tokenizer.encode(test_case)
    assert all([isinstance(x, int) for x in ids])

    assert [sp_tokenizer.vocab[i] for i in tokenized] == ids


def test_cli_functions_e2e(tmp_path):
    """Test we can run on the SP models."""
    prefix = f"{tmp_path}/model"
    cli_utils.sentencepiece_trainer(fixtures.PF01152_PATH_FULL, prefix)
    tokenized_seqs = list(cli_utils.sentencepiece_tokenizer(fixtures.PF01152_PATH_FULL, prefix))
    print(tokenized_seqs)


def test_settings_serialization(tmp_path):
    """Test writing the settings to a file, then getting its results."""
    settings_path = Path(tmp_path) / "settings.json"
    settings = sentence_piece_tokenizer.BioSequencePieceSettings(model_prefix="test")

    settings_path.write_text(json.dumps(settings.dict()))
    settings = json.loads(settings_path.read_text())



# (c) Copyright 2019 Trent Hauck
# All Rights Reserved
"""Module for Sentence Piece tokenization.

This tokenizer uses bindings to call a faster implementation of the algorithm, for now this is the
google library: https://github.com/google/sentencepiece.

This tokenizer needs to be trained to learn its tokenization logic. See the `fit*` methods for
different ways to fit the model prior to use.
"""

import shutil
import tempfile
from pathlib import Path
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from Bio import SeqIO
from pydantic import Field
from typing_extensions import Literal

from gcgc.tokenizer.base import SequenceTokenizer
from gcgc.tokenizer.base import SequenceTokenizerSettings

try:
    import sentencepiece as spm

    # pylint: disable=invalid-name
    has_spm = True
except ImportError:
    # pylint: disable=invalid-name
    has_spm = False


class BioSequencePieceSettings(SequenceTokenizerSettings):
    """The settings for the sentence piece model.

    Like the baseclass, `SequenceTokenizerSettings`, the schema (and thus available fields), can be
    seen by using the `print_schema` classmethod.

    ```python
    >>> print(BioSequencePieceSettings.schema_json(indent=2))
    {
      "title": "SequenceTokenizerSettings"
      ...
    }
    """

    model_prefix: Union[str, Path] = Field(..., env="GCGC_SP_MODEL_PREFIX")
    vocab_size: int = Field(8000, env="GCGC_SP_VOCAB_SIZE")
    model_type: Literal["unigram", "char"] = Field("unigram", env="GCGC_SP_MODEL_TYPE")
    max_sequence_length: int = 4192
    num_threads: int = Field(
        16, description="The number of threads to use.", env="GCGC_SP_NUM_THREADS"
    )
    num_sub_iterations: int = Field(2, env="GCGC_SP_NUM_SUB_ITERATIONS")

    unk_token: Optional[str] = Field("?", env="GCGC_UNK_TOKEN")
    unk_token_id: Optional[int] = Field(0, env="GCGC_UNK_TOKEN_ID")

    @property
    def _model_prefix_path(self) -> Path:
        """Return the prefix as a path not matter its actual type."""
        if isinstance(self.model_prefix, Path):
            return self.model_prefix
        return Path(self.model_prefix)

    @property
    def model_path(self) -> Path:
        """Return the model path based on the prefix."""
        # pylint: disable=no-member
        return self._model_prefix_path.with_suffix(".model")

    @property
    def model_vocab(self) -> Path:
        """Return the model vocab based on the prefix."""
        # pylint: disable=no-member
        return self._model_prefix_path.with_suffix(".vocab")


class BioSequencePiece(SequenceTokenizer):
    """A sentence piece for model on biological sequences."""

    def __init__(self, settings: Optional[BioSequencePieceSettings] = None):
        """Init the BioSequencePiece class.

        Args:
            settings: The settings for the tokenizer.

        """
        if not has_spm or not shutil.which("spm_train"):
            raise RuntimeError("Trying to use sentencepiece but the python library is missing!")

        self.settings = settings or BioSequencePieceSettings()
        super().__init__(settings)

        self.vocab: Dict[str, int] = {}
        self._sp_processor = None

    @property
    def sp_processor(self) -> spm.SentencePieceProcessor:
        """Return the SequencePiece process object."""
        if self._sp_processor is not None:
            return self._sp_processor

        self._sp_processor = spm.SentencePieceProcessor()
        self._sp_processor.load(str(self.settings.model_path))
        return self._sp_processor

    def fit_on_fasta(self, fasta_file: Path):
        """Run the the SP algo on the fasta_file."""
        with tempfile.TemporaryDirectory() as tmpdir:
            tmppath = Path(tmpdir)

            text_file_path = tmppath / "input_textfiles.txt"
            with text_file_path.open("w") as text_lines, fasta_file.open("r") as input_handler:
                for record in SeqIO.parse(input_handler, "fasta"):
                    text_lines.write(f"{str(record.seq)}\n")

            self.fit_on_text(text_file_path)

    def fit_on_list(self, sequence_list: List[str]):
        """Fit the SP algo on a list."""
        with tempfile.TemporaryDirectory() as tmpdir:
            tmppath = Path(tmpdir)

            text_file_path = tmppath / "input_textfiles.txt"
            text_file_path.write_text("\n".join(sequence_list))
            self.fit_on_text(text_file_path)

    def fit_on_text(self, text_file: Path):
        """Run the the SP algo on the text_file."""
        args = [
            f"--input={str(text_file)}",
            f"--model_prefix={self.settings.model_prefix}",
            f"--vocab_size={self.settings.vocab_size}",
            f"--model_type={self.settings.model_type}",
            f"--max_sentence_length={self.settings.max_sequence_length}",
            f"--num_sub_iterations={self.settings.num_sub_iterations}",
            f"--num_threads={self.settings.num_threads}",
            f"--unk_piece={self.settings.unk_token}",
            f"--unk_id={self.settings.unk_token_id}",
        ]

        if self.settings.unk_token:
            args.extend([f"--unk_piece={self.settings.unk_token}"])
        else:
            args.extend(["--unk_id=-1"])

        if self.settings.bos_token:
            args.extend([f"--bos_piece={self.settings.bos_token}"])
        else:
            args.extend(["--bos_id=-1"])

        if self.settings.eos_token:
            args.extend([f"--eos_piece={self.settings.eos_token}"])
        else:
            args.extend(["--eos_id=-1"])

        if self.settings.pad_token:
            args.extend([f"--pad_piece={self.settings.pad_token}", "--pad_id=-1"])
            self.vocab[self.settings.pad_token] = -1
        else:
            args.extend(["--pad_id=-1"])

        spm.SentencePieceTrainer.Train(" ".join(args))

        self.load_vocab()

    def encode(self, seq: str) -> List[int]:
        """Encode the underlying sequence into a list of tokens."""
        return [
            self.vocab.get(s, self.vocab.get(self.settings.unk_token))
            for s in self.encode_as_tokens(seq)
        ]

    def encode_as_tokens(self, seq: str) -> List[str]:
        """Tokenize the sequence into a list of token tokens.

        Args:
            seq: The sequence to encode.

        Returns:
            The list of strs that are the tokens.

        """
        if not self.vocab:
            self.load_vocab()

        return super().apply_length_constraints(self.sp_processor.EncodeAsPieces(seq))

    def load_vocab(self):
        """Load the vocabulary from the file."""
        for line, token in enumerate(self.settings.model_vocab.open()):
            token = token.strip("\n").split("\t")[0]
            self.vocab[token] = line



# (c) Copyright 2020 Trent Hauck
# All Rights Reserved
"""Functions that the CLI commands delegate to."""

import pathlib
from typing import Any
from typing import Dict
from typing import Iterator

from Bio import SeqIO

from gcgc.tokenizer.sentence_piece_tokenizer import BioSequencePiece
from gcgc.tokenizer.sentence_piece_tokenizer import BioSequencePieceSettings


def sentencepiece_trainer(input_fasta: str, model_prefix: str):
    """Train the SP model."""
    input_path = pathlib.Path(input_fasta)
    if not input_path.exists():
        raise ValueError(f"{input_path} does not exist.")

    bsp = BioSequencePiece(settings=BioSequencePieceSettings(model_prefix=model_prefix))
    bsp.fit_on_fasta(input_path)


def sentencepiece_tokenizer(input_fasta: str, model_prefix: str) -> Iterator[Dict[str, Any]]:
    """Process the records and yield the results to the CLI.

    Yields:
        Tokenized records.

    """
    tokenizer = BioSequencePiece(settings=BioSequencePieceSettings(model_prefix=model_prefix))

    with open(input_fasta) as input_handler:
        for record in SeqIO.parse(input_handler, format="fasta"):
            tokenized_sequence = tokenizer.encode(str(record.seq))
            output_record = {
                "token_ids": tokenized_sequence,
                "sequence_id": record.id,
                "sequence_description": record.description,
                "n_tokens": len(tokenized_sequence),
            }

            yield output_record



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""CLI Functions for GCGC."""

import json

import click
from Bio import SeqIO

from gcgc import cli_utils
from gcgc.tokenizer.kmer_tokenzier import KmerTokenizer
from gcgc.tokenizer.kmer_tokenzier import KmerTokenizerSettings


@click.group()
def cli():
    """Console script for gcgc."""


@cli.command()
def version():
    """Print the version and exit."""
    click.echo("0.12.1-dev.5")


@cli.group("tokenizer")
def tokenizer_group():
    """Entrypoint for the tokenizer command."""


@tokenizer_group.command("kmer")
@click.argument("input_fasta", type=str)
@click.option(
    "--vocab_path", type=str, default="kmer_vocab.json", help="Where to save the vocabulary."
)
def kmer_tokernizer(input_fasta: str, vocab_path: str):
    """Run the kmer tokenizer.

    \b
    INPUT_FASTA is the path to the FASTA file to use for training.

    """
    tokenizer = KmerTokenizer(settings=KmerTokenizerSettings())
    with open(input_fasta) as input_handler:
        for record in SeqIO.parse(input_handler, format="fasta"):
            tokenized_sequence = tokenizer.encode(str(record.seq))
            output_record = {
                "token_ids": tokenized_sequence,
                "sequence_id": record.id,
                "sequence_description": record.description,
                "n_tokens": len(tokenized_sequence),
            }
            click.echo(json.dumps(output_record))

    if vocab_path:
        with open(vocab_path, "w") as vocab_file:
            json.dump(tokenizer.vocab, vocab_file, indent=2)


@tokenizer_group.command("sentencepiece")
@click.argument("input_fasta", type=str)
@click.argument("model_prefix", type=str)
def sentencepiece_tokenizer(input_fasta: str, model_prefix: str):
    """Use a pretrained sentencepiece tokenizer to separate a FASTA into a space separate text file.

    See `gcgc train sentencepiece` for information on how to generate the pretrained model.

    \b
    INPUT_FASTA is the path to the FASTA file to use for training.
    MODEL_PREFIX is the path of the pretrained model.

    """
    for output_record in cli_utils.sentencepiece_tokenizer(input_fasta, model_prefix):
        click.echo(json.dumps(output_record))


@cli.group()
def train():
    """Entrypoint for the train command."""


@train.command("sentencepiece")
@click.argument("input_fasta", type=str)
@click.argument("model_prefix", type=str)
def sentencepiece_train(input_fasta: str, model_prefix: str):
    """Use sentencepiece to fit a subword tokenization model.

    See "SentencePiece: A simple and language independent subword tokenizer and detokenizer for
    Neural Text Processing" (Kudo 2018) for more information
    (https://github.com/google/sentencepiece).

    \b
    INPUT_FASTA is the path to the FASTA file to use for training.
    MODEL_PREFIX is the prefix to use for the trained model and the vocabulary.

    """
    cli_utils.sentencepiece_trainer(input_fasta, model_prefix)
